//
// PCC405EP LCD test program
//

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <time.h>
#include <fcntl.h>		// open, close
#include "../../include/ppc405epLCD.h"

int lcdSetSize(int fd, int row, int col);
int lcdInstall(const char driverName[]);

#define LCD_DEVICE_NAME "/dev/lcd"

const struct option longOpt[] =
{
	{"help",		0, NULL, 'h'},
	{"clear",		0, NULL, 'c'},
	{"backlight",	1, NULL, 'l'},
	{"display",		1, NULL, 'd'},
	{"home",		0, NULL, '0'},
	{"position",	1, NULL, 'p'},
	{"cursor",		1, NULL, 'o'},
	{"time",		1, NULL, 't'},
	{"setsize",		1, NULL, 's'},
	{"install",		1, NULL, 'i'},
	{NULL, 0, NULL, 0}
};

const char * shortOpt = "hcl:d:0p:o:ts:i:";	// ':'s are space holders for option arguments

void progHelp(char *progName)
{
	printf("%s [-h] | [option]\n", progName);
	printf("  -h,     --help        : this message (other option ignored)\n"
		   "  -c,     --clear       : clear LCD\n"
		   "  -l o,   --backlight o : back light on (o=1), off (o=0)\n"
		   "  -d str, --display str : display str (use '...' to include spaces)\n"
		   "  -0,     --home        : (number zero) home cursor\n"
		   "  -p pos, --position pos: cursor position = row*100 + column\n"
		   "  -o o,   --cursor o    : (small lettle o) cursor blinking (o > 1),\n"
		   "                          on (o=1), off (o=0)\n"
		   "  -t,     --time        : display current date and time\n"
		   "  -s,     --setsize     : set lcd size, row*100 + column\n"
		   "  -i,     --install path: install LCD driver. path the lcd module path\n"
		   "                          if 0path, module path will be removed\n"
		   "\n");
}

int showDateTime(int lcdFd)
{
	int result;
	struct timeval currTime;

	char dispBuf[17];

	struct tm *pTm;
	char timeStrBuf[27];
	char *pDisp;

	result = lcdClear(lcdFd);
	if (result)
		return result;

	result = lcdCursorOn(lcdFd, 1);	// underline on
	if (result)
		return result;

	gettimeofday(&currTime, NULL);
	pTm = localtime(&currTime.tv_sec);
	strftime(dispBuf, sizeof(dispBuf), "%Y-%m-%d", pTm);
	result = lcdWriteStr(lcdFd, dispBuf);
	if (result <= 0)
	{
		printf("Write date failed\n");
		return result;
	}

	result = lcdSetCursorPosi(lcdFd, 1, 0);
	if (result)
	{
		printf("Set cursor posi failed\n");
		return result;
	}

	strftime(dispBuf, sizeof(dispBuf), "%H:%M:%S", pTm);
	result = lcdWriteStr(lcdFd, dispBuf);
	if (result <= 0)
	{
		printf("Write time failed\n");
	}
	return result;

}

int main(int argc, char* argv[])
{
	int lcdFd;
	char caMajor[10];
	int row, col;
	int nextOpt;
	int idx;
	int result = -1;

	lcdFd = open(LCD_DEVICE_NAME, O_RDWR);
	if (lcdFd < 0)
	{
		printf("Cannot open %s\n", LCD_DEVICE_NAME);
		// do not exit in case of -i
	}

	if (argc == 1)
	{
		progHelp(argv[0]);
	}
	else
	{
		nextOpt = 0;
		idx = 1;
		result = 0;
		do {
			nextOpt = getopt_long( argc, argv, shortOpt, longOpt, NULL);

			switch (nextOpt)
			{
			case 'h':
				progHelp(argv[0]);
				close(lcdFd);
				exit(0);
				break;

			case 'c':
				result = lcdClear(lcdFd);
				printf("Clear result: %d\n", result);
				break;

			case 'l': // small letter l
				if ((optarg != NULL) && (optarg[0] == '1'))
					result = lcdBackLight(lcdFd, 1);
				else
					result = lcdBackLight(lcdFd, 0);
				break;

			case 'd':
				printf("-d opt=%s\n", optarg);
				result = lcdWriteStr(lcdFd, optarg);
				printf("Write Str result: %d\n", result);
				break;

			case '0':	// number zero
				result = lcdHome(lcdFd);
				printf("Home result: %d\n", result);
				break;

			case 'p':
				sscanf(optarg, "%d", &row);
				printf("-p = %d\n", row);
				col = row % 100;
				row /= 100;
				result = lcdSetCursorPosi(lcdFd, row, col);
				printf("Cursor posi (%d, %d)result: %d\n", row, col, result);
				break;

			case 'o':
				sscanf(optarg, "%d", &result);
				result = lcdCursorOn(lcdFd, result);
				printf("Cursor On result: %d\n", result);
				break;

			case 't':
				result = showDateTime(lcdFd);
				break;

			case 's':
				sscanf(optarg, "%d", &row);
				printf("-p = %d\n", row);
				col = row % 100;
				row /= 100;
				result = lcdSetSize(lcdFd, row, col);
				printf("Size (%d, %d)result: %d\n", row, col, result);
				break;

			case 'i':
				if (lcdFd >= 0)
					close(lcdFd);
				result = lcdInstall(optarg);
				lcdFd = open(LCD_DEVICE_NAME, O_RDWR);
				break;

			case -1:
				break;	// no more option

			default:
				abort();
			}

		} while (nextOpt != -1);
	}

	if (lcdFd >= 0)
		close(lcdFd);
	
	return 0;
}
