//
// This file contains PPC405EP LCD user space functions
//

#include <stdio.h>
#include <fcntl.h>		// write(), read()
#include <sys/ioctl.h>	// ioctl()
#include <string.h>

#include "../../include/ppc405epLCD.h"

#define LCD_CLEAR		0x01
#define LCD_HOME		0x02

#define LCD_CURSOR_POSI	0x80

#define LCD_CURSOR_OFF	0x0C

#define LCD_BLINK_OFF	0x0E	// underline cursor on
#define LCD_BLINK_ON	0x0F	// blink and underline

int lcdSetSize(int fd, int row, int col);
int lcdInstall(const char driverName[]);

int lcdSetSize(int fd, int row, int col)
{
	int setRowCol = row * 256 + col;

	return ioctl(fd, IOCTRL_PPC405EP_LCD_SET_SIZE, setRowCol);
}

/*
A script to install device drivers with dynamic assigned major number.
This is from Linux Device Drivers, 2nd Edition

#!/bin/sh
module="scull"
device="scull"
mode="664"

# invoke insmod with all arguments we were passed
# and use a pathname, as newer modutils don't look in . by default
/sbin/insmod -f ./$module.o $* || exit 1

# remove stale nodes
rm -f /dev/${device}[0-3]

major=`awk "\\$2==\"$module\" {print \\$1}" /proc/devices`

mknod /dev/${device}0 c $major 0
mknod /dev/${device}1 c $major 1
mknod /dev/${device}2 c $major 2
mknod /dev/${device}3 c $major 3

# give appropriate group/permissions, and change the group.
# Not all distributions have staff; some have "wheel" instead.
group="staff"
grep '^staff:' /etc/group > /dev/null || group="wheel"

chgrp $group /dev/${device}[0-3]
chmod $mode /dev/${device}[0-3]
*/
int lcdInstall(const char driverName[])
{
#define LCD_DEVICE_NAME "/dev/lcd"

	int lcdFd;
	char sysCmd[81];
	int retVal;

	printf("LCD device driver: %s\n", driverName);
	sleep(1);

	if (driverName[0] == '0')
	{
		sprintf(sysCmd, "rm %s", LCD_DEVICE_NAME);
		printf("system: %s\n", sysCmd);
		system(sysCmd);

		sprintf(sysCmd, "rmmod %s", driverName+1);
		printf("system: %s\n", sysCmd);
		system(sysCmd);
		return 0;
	}

	lcdFd = open(LCD_DEVICE_NAME, O_RDWR);
	if (lcdFd >= 0)
	{ // /dev/lcd exists
		close(lcdFd);
		return 0;
	}

	if (strlen(driverName) > 30)
		return -1;

	strcpy(sysCmd, "insmod ");
	strcat(sysCmd, driverName);
	retVal = system(sysCmd);
	printf("Insmod = %d\n", retVal);
	if (retVal != 0)
		return retVal;

	lcdFd = open("/proc/lcdmodmajor", O_RDONLY);
	if (lcdFd >= 0)
	{
		read(lcdFd, sysCmd, 10);
		sscanf(sysCmd, "%d", &retVal);
		printf("LCD Module Major # = %d\n", retVal);
		close(lcdFd);

		sprintf(sysCmd, "mknod %s c %d 0", LCD_DEVICE_NAME, retVal);
		retVal = system(sysCmd);
		printf("mknod = %d\n", retVal);
	}
	else
		return -2;

	return retVal;
}

int lcdClear(int fd)
{
	int ret;
	int retries = 3;

	do {
		ret = ioctl(fd, IOCTRL_PPC405EP_LCD_CMD, LCD_CLEAR);
		usleep(2000);	// although driver will check, but less do it here anyway
	} while ((ret != 0) && (retries-- > 0));

	return ret;
}

int lcdCmd(int fd, char cmd)
{
	int ret;
	int retries = 3;

	do {
		ret = ioctl(fd, IOCTRL_PPC405EP_LCD_CMD, cmd);
		if (ret != 0)
			usleep(50);
	} while ((ret != 0) && (retries-- > 0));

	return ret;
}

int lcdData(int fd, char data)
{
	int ret;
	int retries = 3;

	do {
		ret = ioctl(fd, IOCTRL_PPC405EP_LCD_DATA, data);
		if (ret != 0)
			usleep(50);
	} while ((ret != 0) && (retries-- > 0));

	return ret;
}

int lcdHome(int fd)
{
	return lcdCmd(fd, LCD_HOME);
}

int lcdSetCursorPosi(int fd, char row, char col)
{
	char posi;
	int ret;

	posi = LCD_CURSOR_POSI | (row * 0x40 + col);
	return lcdCmd(fd, posi);
}

int lcdCursorOn(int fd, int on)
{
#define LCD_BLINK_OFF	0x0E	// underline cursor on
#define LCD_BLINK_ON	0x0F	// blink and underline

	char onOff;

	switch (on)
	{
	case 0:
		onOff = LCD_CURSOR_OFF;
		break;

	case 1: 
		onOff = LCD_BLINK_OFF;
		break;

	default:
		onOff = LCD_BLINK_ON;
		break;
	}

	return lcdCmd(fd, onOff);
}

int lcdBackLight(int fd, int on)
{
	int ret;
	int retries = 3;

	do {
		ret = ioctl(fd, IOCTRL_PPC405EP_LCD_SET_LIGHT, on);
		usleep(50);	// although driver will check, but less do it here anyway
	} while ((ret != 0) && (retries-- > 0));

	return ret;
}

int lcdWriteChar(int fd, char cV)
{
	int ret;
	int retries = 3;

	do {
		ret = write(fd, &cV, 1);
		usleep(50);
	} while ((ret != 1) && (retries-- > 0));

	if (ret != 1)
	{
		printf("Write Char error: %d\n", ret);
	}

	return ret;
}

int lcdWriteStr(int fd, char *str)
{
	int ret, len;
	int idx = 0;
	
	if (str == (char *) 0)
	{
		return -EINVAL;
	}

	len = strlen(str);
	for (idx = 0; idx < len; idx++)
	{
		ret = lcdWriteChar(fd, str[idx]);
		if (ret != 1)
			break;
	}

	return idx;
}
