/*
 * This is LCD module for PCC 405EP
 * Author: Charles Lee
 *         WELCM <charles@welcm.com>
 *
 * 2007 (c) WELCM.  This file is licensed under the terms of 
 * the GNU General Public License. This program is licensed "as is"
 * without any warranty of any kind, whether express or implied.
 */

//
// This is LCD module for PCC 405EP
// To avoid potential kernel namespace conflict, this module is
// coded in one file. Hopefully it will not be too big
//
// LCD devices may require some "black-out" period, such as clean screen and reset.
// They also have relative longer access time, 50 usec. 
//
// One design decision is to make any delay in this module as short as possible
// to allow other kernel related functions, such as 1PPS interrupt from (GPS) clock
// device, response faster.
//

#include <linux/module.h>
#include <linux/kernel.h>
#include <asm/io.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <linux/delay.h>

#include "../../include/ppc405epLCD.h"

//#define DEVICE_NAME "lcd"
#define PPC405EP_LCD_DEVICE_NAME	"lcd405ep"

#define SUCCESS		0

static int modMajor = 0;		// 0 - for dynamic Major # allocation
static int deviceOpened = 0;

static int lcdRows = 2;
static int lcdCols = 16;

static u64 lastJiffies = 0;
static u64 waitJiffies = 0;

static char addr_flag = 0x80;

#define CPLD_REG0_ADDR 0x50100000

static void __iomem *cpld_virt;
#define LCD_BLK_CTRL  (volatile char *)((ulong)cpld_virt+1)
#define LCD_CMD_ADDR  (volatile char *)((ulong)cpld_virt+2)
#define LCD_DATA_ADDR (volatile char *)((ulong)cpld_virt+3)

static int initLCDMod(void);
static void exitLCDMod(void);

static int deviceOpen(struct inode *, struct file *);
static int deviceRelease(struct inode *, struct file *);	// close

static int deviceIOCtl(struct inode *, struct file *, unsigned int, unsigned long);
static loff_t deviceSeek(struct file*, loff_t, int);

static ssize_t deviceRead(struct file*, char *, size_t, loff_t *);
static ssize_t deviceWrite(struct file*, const char *, size_t, loff_t *);

static struct file_operations fops =
{
	.open = deviceOpen,

	.ioctl = deviceIOCtl,
	.llseek = deviceSeek,
	.read = deviceRead,
	.write = deviceWrite,

	.release = deviceRelease
};

#define PROC_FILE_NAME "lcdmodmajor"
static struct proc_dir_entry *procFile;

static int procFileRead(char * buffer, char **bufLoc,
						off_t offset, int bufLen, int *eof, void* data)
{
	int ret;

	if (offset > 0)
	{
		ret = 0;
	}
	else
	{
		ret = sprintf(buffer, "%d\n", modMajor);
	}
	return ret;
}

static int isBusy(void)
{
	u64 currJiffies;

	if (waitJiffies > 0)
	{
		currJiffies = get_jiffies_64();
		if (currJiffies < lastJiffies)
		{ // overflow occured
			lastJiffies = 0;	// assume last is 0
			waitJiffies --;		// one jiffy passed at least
			return 1;
		}
		if ((currJiffies - lastJiffies) < waitJiffies)
			return 1;	// need to wait more

		waitJiffies = 0;	// no more wait
	}

	if(*LCD_CMD_ADDR & 0x80)
	{
		return 1;
	}

	return 0;	// ready
}

static int deviceOpen(struct inode *pINode, struct file *pF)
{
	if (deviceOpened)
		return -EBUSY;

	deviceOpened = 1;

	try_module_get(THIS_MODULE);	// increment reference count to prevent rmmod

	cpld_virt = ioremap((ulong)CPLD_REG0_ADDR, 4);

	return SUCCESS;
}

static int deviceRelease(struct inode * pINode, struct file *pF)	// close
{
	deviceOpened = 0;

	iounmap((void*)cpld_virt);

	module_put(THIS_MODULE);	// decrement reference count to allow rmmod
	return SUCCESS;
}

static int deviceIOCtl(struct inode *pINode, struct file * pF,
					   unsigned int request, unsigned long ulParm)
{
	char cVal;
	int aiParm[4];

	switch (request)
	{
	case IOCTRL_PPC405EP_LCD_SET_SIZE:
		aiParm[0] = ulParm / 256;
		aiParm[1] = ulParm % 256;

		printk(KERN_ALERT " in set size, %ld, %d,%d\n", ulParm, aiParm[0], aiParm[1]);

		if ((aiParm[0] > 0) && (aiParm[0] < 11)
			&&
			(aiParm[1] > 7) && (aiParm[1] < 81))
		{
			lcdRows = aiParm[0];
			lcdCols = aiParm[1];
		}
		else
		{
			return -EINVAL;
		}

		break;

	case IOCTRL_PPC405EP_LCD_GET_SIZE:
		return (lcdRows << 8) + lcdCols;
		break;

	case IOCTRL_PPC405EP_LCD_CMD:
		if (isBusy())
			return -EBUSY;

		*LCD_CMD_ADDR = (char) ulParm;
		if (ulParm == 1)
		{ // clear LCD, 2 ms wait needed
			lastJiffies = get_jiffies_64();
			waitJiffies = 2000*HZ/1000000+1;
		}
		break;

	case IOCTRL_PPC405EP_LCD_DATA:
		if (isBusy())
			return -EBUSY;

		*LCD_DATA_ADDR = (char) ulParm;
		break;

	case IOCTRL_PPC405EP_LCD_SET_LIGHT:
		cVal = *LCD_BLK_CTRL;
		if (ulParm)		// turn backlight on
		{
			*LCD_BLK_CTRL |= 0x2;
		}
		else
		{
			*LCD_BLK_CTRL &= (~0x2);
		}

		break;

	default:
		return -EINVAL;	// unknown request code
	}
	return SUCCESS;
}

static loff_t deviceSeek(struct file* pF, loff_t offset, int val)
{
	return 0;
}


static ssize_t deviceRead(struct file *pF, char __user *buf,
						  size_t length, loff_t *pOffset)
{
	return -EINVAL;
}

// LCD device is assumed in no blocking mode
static ssize_t deviceWrite(struct file*pF, const __user char *buf,
						   size_t length, loff_t *pOffset)
{
	char addr;
	char val;

	if (buf == (char*) 0)
	{
		printk(KERN_ALERT "Write NULL\n");

		return -EINVAL;
	}

	if (isBusy())
	{
		printk(KERN_ALERT "Write busy\n");
		return -EAGAIN;
	}

	addr = *LCD_CMD_ADDR;	// get next character position
	udelay(50);
	addr &= 0x1F;	// get position on row (most 32 characters/row)
					// this is to avoid setting cursor position at row 2 col 0,
					// which addr is 0x40 and
	if ((addr != 0) && ((addr % lcdCols) == 0))
	{
		addr_flag ^= 0x40;	// switch lines
		*LCD_CMD_ADDR = addr_flag;
		udelay(50);
	}

	get_user(val, buf);

	*LCD_DATA_ADDR = val;
	udelay(50);

	return 1;	// one character/per call
}

static int initLCDMod(void)
{
	modMajor = register_chrdev(modMajor, PPC405EP_LCD_DEVICE_NAME, &fops);
	if (modMajor == 0)
		return -ENOMEM;
	
	printk(KERN_ALERT "lcd module inserted: %d\n", modMajor);

	procFile = create_proc_entry(PROC_FILE_NAME, 0644, &proc_root);
	if (procFile == NULL)
	{
		remove_proc_entry(PROC_FILE_NAME, &proc_root);
		unregister_chrdev(modMajor, PPC405EP_LCD_DEVICE_NAME);

		return -ENOMEM;
	}

	procFile->read_proc = procFileRead;
	procFile->owner = THIS_MODULE;
	procFile->uid = 0;
	procFile->gid = 0;
	procFile->size = 10;

	return modMajor;
}

static void exitLCDMod(void)
{
	int ret;
	
	remove_proc_entry(PROC_FILE_NAME, &proc_root);
	ret = unregister_chrdev(modMajor, PPC405EP_LCD_DEVICE_NAME);
	if (ret < 0)
		printk(KERN_ALERT "lcd module removed with error: %d\n", ret);
}

module_init(initLCDMod);
module_exit(exitLCDMod);

// Specify licence, author, description, and suppoted device
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Charles Lee @ charles@welcm.com");
MODULE_DESCRIPTION("LCD driver");
MODULE_SUPPORTED_DEVICE("LCD - AMCC Taihu board (PPC405EP)");
